import threading as _threading

from .exceptions import SkipFixerException
from .registry import PatchingRegistry, MultiPatchingRegistry
from .runner import PatchingRunner
from .utilities import (
    PatchingUtilities,
    WarningsProxy,
    tuplify_software_version,
    detuplify_software_version,
)


def generic_patch_software(
    settings,
    patching_registry,
    patching_utilities_class=PatchingUtilities,
    patching_runner_class=PatchingRunner,
    warnings_proxy=None,
):
    """Load all dependencies, and apply relevant fixers from the `patching_registry`,
    according to the settings of the provided `settings`.

    You can provide custom classes to be instantiated instead of default ones, and/or an
    existing WarningsProxy which will be updated with the new settings as soon as possible.
    """

    patching_registry.populate()
    assert patching_registry._is_populated

    patching_utilities = patching_utilities_class(settings=settings)

    if warnings_proxy:  # Update the config of preexisting WarningsProxy
        warnings_proxy.set_patching_utilities(patching_utilities)

    patching_runner = patching_runner_class(
        settings=settings,
        patching_utilities=patching_utilities,
        patching_registry=patching_registry,
    )
    patching_runner.patch_software()

    return None  # No return value for now, if some is set one day it shall be a DICT!


#: Example configuration to copy() and adapt
DEFAULT_SETTINGS = dict(
    logging_level="INFO",
    enable_warnings=True,
    patch_injected_objects=True,
    include_fixer_ids="*",
    include_fixer_families=None,
    exclude_fixer_ids=None,
    exclude_fixer_families=None,
)


#: Lock meant to globally protect all the patching workflow
PATCHING_LOCK = _threading.RLock()


def make_safe_patcher(f):
    """
    This decorator makes a patching launcher thread-safe with a recursive lock.

    Other checks and misc. features might be added in the future, so packages using
    this patching framework should always decorate their main "patch()" entrypoint
    with this utility.
    """
    import functools

    @functools.wraps(f)
    def inner(*args, **kwargs):
        with PATCHING_LOCK:
            return f(*args, **kwargs)

    return inner



from __future__ import absolute_import, print_function, unicode_literals

from compat_patcher_core.registry import PatchingRegistry


def populate_fixers_registry(registry):
    from . import fixers


# TODO fetch and return the real version of the patched framework/library
def get_current_software_version():
    return "1.2.3"


# This must be instantiated HERE so that fixer submodules can access it at import time
patching_registry = PatchingRegistry(
    family_prefix="{{ cookiecutter.project_prefix }}",
    populate_callable=populate_fixers_registry,
    current_software_version=get_current_software_version,
)

register_compatibility_fixer = (
    patching_registry.register_compatibility_fixer
)  # Shortcut












from __future__ import absolute_import, print_function, unicode_literals

import collections
import itertools

import six

from compat_patcher_core.utilities import (
    tuplify_software_version,
    _import_attribute_from_dotted_string,
)


class PatchingRegistry(object):
    """
    This registry is used to store and select a set of fixers related to some
    specific software.

    `family_prefix` will be used to constructe family names, along with the software
    reference version provided by the fixer.

    `populate_callable`, if provided, is a callable taking the registry as first
    argument, and which will be called by `populate()`.

    `current_software_version` may be a version tuple or a string. If it's None,
    then an override value will have to be provided when calling `get_relevant_fixers`.
    """

    def __init__(
        self, family_prefix, populate_callable=None, current_software_version=None
    ):
        assert family_prefix and isinstance(
            family_prefix, six.string_types
        ), family_prefix
        assert populate_callable is None or hasattr(
            populate_callable, "__call__"
        ), populate_callable
        self._family_prefix = family_prefix
        self._is_populated = False
        self._populate_callable = populate_callable
        self._patching_registry = collections.OrderedDict()
        self._current_software_version = current_software_version

    def _get_current_software_version(self):
        """
        Returns a tuple of integers, or a dotted string, representing the current
        version of the software to be patched.
        """
        current_software_version = self._current_software_version
        if six.callable(current_software_version):
            current_software_version = current_software_version()
        assert current_software_version is None or isinstance(
            current_software_version, (six.string_types, tuple, list)
        ), current_software_version
        return current_software_version

    def populate(self):
        """
        Trigger the registration of potential lazy fixers, which might be in other
        submodules, or waiting in factory functions.
        """
        res = None
        if not self._is_populated:
            if self._populate_callable:
                res = self._populate_callable(self)
            self._is_populated = True
        return res

    @staticmethod
    def _extract_docstring(func):
        """Extract and check the docstring of a callable"""
        doc = func.__doc__
        if not doc:
            raise ValueError("Fixer %r must provide a help string" % func)
        return doc

    def register_compatibility_fixer(
        self,
        fixer_reference_version,
        fixer_applied_from_version=None,  # INCLUDING this version
        fixer_applied_upto_version=None,  # EXCLUDING this version
        feature_supported_from_version=None,
        feature_supported_upto_version=None,
        fixer_tags=None,
    ):
        """
        Register a compatibility fixer, which will be activated only if current
        software version is >= `fixer_applied_from_version` and <
        `fixer_applied_upto_version` (let them be None to have no limit).

        The "fixer_reference_version" parameters identifies the software version
        where the breaking change was introduced (for backwards compatibility
        fixers), or where the new feature was introduced (for forwards compatibility
        fixers). It is not related to the appearance of corresponding
        DeprecationWarnings in the software. It is also used to sort fixers when
        applying them, and to generate the name of the family of fixers concerned.

        `feature_supported_from_version` (included) and
        `feature_supported_upto_version` (excluded) may be used to limit the range of
        software versions for which related unit-tests are expected to work (i.e
        versions for which the feature is available, either as a monkey-paching or as
        standard code).

        Version identifiers must be dotted strings, eg. "1.9.1". None means "no limit"
        here.

        `fixer_tags` is a **list** of strings, which can be used to differentiate fixers
        which will be applied at different moments of software startup.
        """

        assert (
            isinstance(fixer_reference_version, six.string_types)
            and fixer_reference_version
        ), fixer_reference_version  # eg. "1.9"
        assert fixer_tags is None or isinstance(fixer_tags, list), fixer_tags

        fixer_family = self._family_prefix + fixer_reference_version
        fixer_reference_version = tuplify_software_version(fixer_reference_version)
        fixer_applied_from_version = tuplify_software_version(
            fixer_applied_from_version
        )
        fixer_applied_upto_version = tuplify_software_version(
            fixer_applied_upto_version
        )
        feature_supported_from_version = tuplify_software_version(
            feature_supported_from_version
        )
        feature_supported_upto_version = tuplify_software_version(
            feature_supported_upto_version
        )
        fixer_tags = fixer_tags or []

        if fixer_applied_from_version and fixer_applied_upto_version:
            assert fixer_applied_from_version < fixer_applied_upto_version
        if feature_supported_from_version and feature_supported_upto_version:
            assert feature_supported_from_version < feature_supported_upto_version

        def _register_simple_fixer(func):
            fixer_id = func.__name__  # untouched ATM, not fully qualified
            new_fixer = dict(
                fixer_callable=func,
                fixer_id=fixer_id,
                fixer_explanation=self._extract_docstring(func),
                fixer_reference_version=fixer_reference_version,
                fixer_family=fixer_family,
                fixer_tags=fixer_tags,
                fixer_applied_from_version=fixer_applied_from_version,
                fixer_applied_upto_version=fixer_applied_upto_version,
                feature_supported_from_version=feature_supported_from_version,
                feature_supported_upto_version=feature_supported_upto_version,
                fixer_qualified_name="%s|%s" % (fixer_family, fixer_id),
            )

            assert fixer_id not in self._patching_registry, (
                "duplicate fixer id %s detected" % fixer_id
            )
            self._patching_registry[fixer_id] = new_fixer
            # print("patching_registry", patching_registry)
            return func

        return _register_simple_fixer

    def get_all_fixers(self):
        """Return the list of all fixers (as dicts) known by this registry."""
        return list(self._patching_registry.values())

    def get_fixer_by_id(self, fixer_id):
        """Return the fixer having this (unqualified) ID, or raise KeyError."""
        return self._patching_registry[fixer_id]

    def get_relevant_fixers(
        self,
        include_fixer_ids="*",
        include_fixer_families=None,
        exclude_fixer_ids=None,
        exclude_fixer_families=None,
        current_software_version=None,
        log=None,
    ):
        """
        Return the list of fixers (as dicts) to be applied for the target software
        version, based on the metadata of fixers, as well as inclusion/exclusion
        lists provided as arguments.

        For inclusion/exclusion filters, a special "*" value means "all fixers",
        else a list of strings is expected.

        An output callable `log` may be provided, expecting a string as argument,
        to debug the reasons why some fixers weren't selected.

        This method forces a populate() on the registry.
        """

        if not current_software_version:
            current_software_version = self._get_current_software_version()
        if not current_software_version:
            raise ValueError(
                "PatchingRegistry must be provided a valid current_software_version "
                "in its constructor or as `get_relevant_fixers` argument."
            )

        self.populate()

        ALL = "*"

        log = log or (lambda x: x)

        current_software_version = tuplify_software_version(current_software_version)

        relevant_fixers = []

        # Shortcut for the common case "no specific inclusion/exclusion lists"
        mass_include = (
            include_fixer_ids == ALL or include_fixer_families == ALL
        ) and not any((exclude_fixer_ids, exclude_fixer_families))

        for fixer_id, fixer in self._patching_registry.items():
            assert fixer_id == fixer["fixer_id"], fixer
            fixer_qualified_name = fixer["fixer_qualified_name"]

            if (
                fixer["fixer_applied_from_version"] is not None
                and current_software_version < fixer["fixer_applied_from_version"]
            ):  # STRICTLY SMALLER
                log(
                    "Skipping fixer %s, useful only in next software versions"
                    % fixer_id
                )
                continue

            if (
                fixer["fixer_applied_upto_version"] is not None
                and current_software_version >= fixer["fixer_applied_upto_version"]
            ):  # GREATER OR EQUAL
                log(
                    "Skipping fixer %s, useful only in previous software versions"
                    % fixer_id
                )
                continue

            if not mass_include:

                included = False
                if include_fixer_ids == ALL or (
                    include_fixer_ids
                    and (
                        fixer_id in include_fixer_ids
                        or fixer_qualified_name in include_fixer_ids
                    )
                ):
                    included = True
                if include_fixer_families == ALL or (
                    include_fixer_families
                    and fixer["fixer_family"] in include_fixer_families
                ):
                    included = True

                if not included:
                    log(
                        "Skipping fixer %s, having neither id nor family (%s) included by patcher settings"
                        % (fixer_id, fixer["fixer_family"])
                    )
                    continue

                if exclude_fixer_ids == ALL or (
                    exclude_fixer_ids
                    and (
                        fixer_id in exclude_fixer_ids
                        or fixer_qualified_name in exclude_fixer_ids
                    )
                ):
                    log("Skipping fixer %s, excluded by patcher settings" % fixer_id)
                    continue

                if exclude_fixer_families == ALL or (
                    exclude_fixer_families
                    and fixer["fixer_family"] in exclude_fixer_families
                ):
                    log(
                        "Skipping fixer %s, having family %s excluded by patcher settings"
                        % (fixer_id, fixer["fixer_family"])
                    )
                    continue

            # cheers, this fixer has passed all filters!
            relevant_fixers.append(fixer)

        return relevant_fixers

    def get_relevant_fixer_ids(self, qualified=False, **kwargs):
        """"Same as `get_relevant_fixers`, but only returns IDs of selected fixers.

        If `qualified` is True, returns a fixers IDs dot-prefixed with the family name."""
        fixers = self.get_relevant_fixers(**kwargs)
        field_name = "fixer_qualified_name" if qualified else "fixer_id"
        return [f[field_name] for f in fixers]


class MultiPatchingRegistry(object):
    """
    This patching registry wraps a list of other registries, each having its own
    fixers and current software version.

    It concatenates and returns selected fixers on demand, assuming that they are
    compatible with each other.
    """

    def __init__(self, registries):
        self._registry_references = registries
        self._is_populated = False
        self._registries = self._load_registries(registries)

    def populate(self):
        res = []
        if not self._is_populated:
            for registry in self._registries:
                res.append(registry.populate())
            self._is_populated = True
        return res

    @staticmethod
    def _load_registries(registry_references):
        registries = []
        for registry_reference in registry_references:

            original_registry_reference = registry_reference

            if isinstance(registry_reference, six.string_types):
                registry_reference = _import_attribute_from_dotted_string(
                    registry_reference
                )

            if six.callable(registry_reference):
                registry_reference = registry_reference()

            if not isinstance(registry_reference, PatchingRegistry):
                raise ValueError(
                    "Wrong registry reference %r" % original_registry_reference
                )

            registries.append(registry_reference)
        assert len(registries) == len(registry_references)
        return registries

    @staticmethod
    def _flatten(list_of_lists):
        return list(itertools.chain(*list_of_lists))

    def get_relevant_fixers(self, *args, **kwargs):
        """Populate underlying registries, and return the concatenation of their
        selected fixers.

        Forcing a `current_software_version` as parameter of this method is still
        possible, but beware that underlying registries all deal with the same
        software stack, in this case.
        """

        self.populate()

        return self._flatten(
            registry.get_relevant_fixers(*args, **kwargs)
            for registry in self._registries
        )

    def get_all_fixers(self, *args, **kwargs):
        """Return the concatenation of all fixers of underlying registries."""
        return self._flatten(
            [registry.get_all_fixers(*args, **kwargs) for registry in self._registries]
        )

    def get_fixer_by_id(self, fixer_id, *args, **kwargs):
        """
        In case of duplicate fixers having the same ID, just return the first one.
        """
        for registry in self._registries:
            try:
                return registry.get_fixer_by_id(fixer_id, *args, **kwargs)
            except KeyError:
                pass
        raise KeyError("Fixer %r not found in any patching registries" % fixer_id)

    get_relevant_fixer_ids = six.get_unbound_function(
        PatchingRegistry.get_relevant_fixer_ids
    )  # Unmodified



from __future__ import absolute_import, print_function, unicode_literals

# Use this this to preconfigure "register_compatibility_fixer" for a set of fixers
from functools import partial

from ..deprecation import *
from ..registry import register_compatibility_fixer


@register_compatibility_fixer(
        fixer_reference_version="1.0",
        fixer_applied_from_version="1.0"
)
def fix_behaviour_submodule_element(utils):
    """Help string is mandatory to explain what the fixer does! Here, something..."""
    utils.emit_warning("Something was deprecated in software", DeprecationWarning)
    utils.inject_import_alias(alias_name="my_import_alias_for_csv", real_name="csv")
    # HERE use injectors from utils to modify the software



import dummy_module
from compat_patcher_core import (
    generic_patch_software,
    PatchingRegistry,
    DEFAULT_SETTINGS,
    make_safe_patcher,
)
from compat_patcher_core.registry import MultiPatchingRegistry
from compat_patcher_core.runner import PatchingRunner
from compat_patcher_core.utilities import PatchingUtilities, WarningsProxy
from dummy_fixers import patching_registry, patching_registry_bis


def test_runner_patch_software():

    PatchingRunner._clear_all_applied_fixers()  # Important

    del dummy_module.APPLIED_FIXERS[:]
    settings = DEFAULT_SETTINGS.copy()

    patching_utilities = PatchingUtilities(settings=settings)

    patching_runner = PatchingRunner(
        settings=settings,
        patching_utilities=patching_utilities,
        patching_registry=patching_registry,
    )

    _sorted_fixers_ids = [
        f["fixer_id"] for f in patching_runner._get_sorted_relevant_fixers()
    ]
    assert (
        "fix_something_but_skipped" in _sorted_fixers_ids
    )  # Will be aborted below during execution of fixers

    result = patching_runner.patch_software()
    # Fixers sorted by descending reference version and name
    assert result == dict(
        fixers_just_applied=[
            "fix_something_upto_v6",
            "fix_something_from_v5",
            "fix_something_always",
            "fix_something_from_v4",
        ]
    )

    assert (
        dummy_module.APPLIED_FIXERS == result["fixers_just_applied"]
    )  # Fixers are really run

    result = patching_runner.patch_software()
    assert result == dict(fixers_just_applied=[])  # Already applied so skipped

    patching_runner2 = PatchingRunner(
        settings=settings,
        patching_utilities=patching_utilities,
        patching_registry=patching_registry,
    )
    result = patching_runner2.patch_software()
    assert result == dict(
        fixers_just_applied=[]
    )  # Already applied so skipped, despite different runner instance


def test_generic_patch_software():

    PatchingRunner._clear_all_applied_fixers()  # Important

    del dummy_module.APPLIED_FIXERS[:]
    settings = DEFAULT_SETTINGS.copy()

    patching_registry_internal = PatchingRegistry(
        family_prefix="internal", current_software_version=lambda: "100"
    )

    @patching_registry_internal.register_compatibility_fixer(
        fixer_reference_version="8.3"
    )
    def fix_stuffs_internal(utils):
        "Does something still"
        assert utils.emit_log  # proper utility
        dummy_module.APPLIED_FIXERS.append(fix_stuffs_internal.__name__)

    warnings_proxy = WarningsProxy()
    assert not warnings_proxy._patching_utilities
    assert not patching_registry_internal._is_populated

    generic_patch_software(
        settings=settings,
        patching_registry=patching_registry_internal,
        warnings_proxy=warnings_proxy,
    )

    assert warnings_proxy._patching_utilities
    assert patching_registry_internal._is_populated
    assert dummy_module.APPLIED_FIXERS == ["fix_stuffs_internal"]


def test_fixer_idempotence_through_runner():

    PatchingRunner._clear_all_applied_fixers()  # Important

    del dummy_module.APPLIED_FIXERS[:]
    settings = DEFAULT_SETTINGS.copy()

    patching_utilities = PatchingUtilities(settings=settings)

    multi_registry = MultiPatchingRegistry(
        registries=[patching_registry, patching_registry_bis]
    )
    relevant_fixers = multi_registry.get_relevant_fixer_ids()
    assert relevant_fixers == [
        "fix_something_from_v4",
        "fix_something_from_v5",
        "fix_something_upto_v6",
        "fix_something_always",
        "fix_something_but_skipped",
        "fix_something_other_taken",
        "fix_something_always",  # This one is TWICE there, under different registries
    ]

    patching_runner2 = PatchingRunner(
        settings=settings,
        patching_utilities=patching_utilities,
        patching_registry=multi_registry,
    )
    result = patching_runner2.patch_software()
    assert result == dict(
        fixers_just_applied=[
            "fix_something_always",
            "fix_something_other_taken",
            "fix_something_upto_v6",
            "fix_something_from_v5",
            "fix_something_always",  # Properly run even if duplicate fixer_id
            "fix_something_from_v4",
        ]
    )


def test_make_safe_patcher():
    import time, threading

    shared_value = [0]

    @make_safe_patcher
    def slow_func(myattr):
        """
        This is a slooooow function.
        """
        value = shared_value[0]
        time.sleep(0.01)
        shared_value[0] = value + 1

    assert "slooooow" in slow_func.__doc__  # Properly wrapped

    threads = [
        threading.Thread(target=slow_func, kwargs=dict(myattr=22)) for i in range(5)
    ]
    [t.start() for t in threads]
    [t.join() for t in threads]

    assert shared_value[0] == 5



from __future__ import absolute_import, print_function, unicode_literals

import functools

import six

from compat_patcher_core.exceptions import SkipFixerException


class PatchingRunner(object):
    """
    This class is in charge of fetching relevant fixers from the registry,
    and applying them in a proper order, while skipping those who have already been
    applied (i.e same family name and ID) and those which raised SkipFixerException.
    """

    # Helpful for autodocumentation
    settings_keys_used = [
        "include_fixer_ids",
        "include_fixer_families",
        "exclude_fixer_ids",
        "exclude_fixer_families",
    ]

    _all_applied_fixers = []  # Class attribute with qualified fixer names!

    def __init__(self, settings, patching_registry, patching_utilities):
        assert settings, settings
        self._settings = settings
        self._patching_registry = patching_registry
        self._patching_utilities = patching_utilities

    @classmethod
    def _clear_all_applied_fixers(cls):  # For testing only!
        del cls._all_applied_fixers[:]

    def _get_patcher_setting(self, name):
        """
        Returns the value of a patcher setting.
        """
        assert name in self.settings_keys_used  # To track coherence

        value = self._settings[name]

        # For now, patching utilities validate their own settings, so we just check filters
        if name.startswith("include") or name.startswith("exclude"):
            assert value in ("*", None) or (
                isinstance(value, (list, tuple))
                and all(isinstance(f, six.string_types) for f in value)
            ), value

        return value

    def _apply_selected_fixers(self, fixers):
        fixers_just_applied = []
        for fixer in fixers:

            fixer_qualified_name = fixer["fixer_qualified_name"]

            if fixer_qualified_name not in self._all_applied_fixers:
                self._patching_utilities.emit_log(
                    "Compat fixer {}->{} is getting applied".format(
                        fixer["fixer_family"], fixer["fixer_id"]
                    ),
                    level="INFO",
                )
                try:
                    fixer["fixer_callable"](self._patching_utilities)
                    self._all_applied_fixers.append(fixer_qualified_name)
                    fixers_just_applied.append(fixer["fixer_id"])
                except SkipFixerException as e:
                    self._patching_utilities.emit_log(
                        "Compat fixer {}->{} was actually not applied, reason: {}".format(
                            fixer["fixer_family"], fixer["fixer_id"], e
                        ),
                        level="WARNING",
                    )
            else:
                self._patching_utilities.emit_log(
                    "Compat fixer {}->{} was already applied".format(["fixer_family"], fixer["fixer_id"]),
                    level="WARNING",
                )
        return fixers_just_applied

    def _get_sorted_relevant_fixers(self):

        # For now, we don't need to be able to force-send a `current_software_version`
        fixers_settings = dict(
            include_fixer_ids=self._get_patcher_setting("include_fixer_ids"),
            include_fixer_families=self._get_patcher_setting("include_fixer_families"),
            exclude_fixer_ids=self._get_patcher_setting("exclude_fixer_ids"),
            exclude_fixer_families=self._get_patcher_setting("exclude_fixer_families"),
        )
        log = functools.partial(self._patching_utilities.emit_log, level="DEBUG")
        relevant_fixers = self._patching_registry.get_relevant_fixers(
            log=log, **fixers_settings
        )

        # REVERSED order is necessary for backwards compatibility, more advanced
        # sorting might be introduced to help forward-compatibility fixers...
        relevant_fixers.sort(
            key=lambda x: (x["fixer_reference_version"], x["fixer_id"]), reverse=True
        )

        return relevant_fixers

    def patch_software(self):
        """Patch the software according to plans.

        Return a dict with, at least field "fixers_just_applied", the list of fixers that
        were successfully applied during this call.
        """
        relevant_fixers = self._get_sorted_relevant_fixers()
        fixers_just_applied = self._apply_selected_fixers(relevant_fixers)
        return dict(fixers_just_applied=fixers_just_applied)



from __future__ import absolute_import, print_function, unicode_literals

from functools import reduce
from io import open

from compat_patcher_core.utilities import detuplify_software_version


def make_table(grid):
    """
    Borrowed from http://stackoverflow.com/a/12539081/5088990
    :param grid: List of lists
    :return:
    """

    def _table_div(num_cols, col_width, header_flag):
        if header_flag == 1:
            return num_cols * ("+" + (col_width) * "=") + "+\n"
        else:
            return num_cols * ("+" + (col_width) * "-") + "+\n"

    def _normalize_cell(string, length):
        return string + ((length - len(string)) * " ")

    cell_width = 2 + max(
        reduce(lambda x, y: x + y, [[len(item) for item in row] for row in grid], [])
    )
    num_cols = len(grid[0])
    rst = _table_div(num_cols, cell_width, 0)
    header_flag = 1
    for row in grid:
        rst = (
            rst
            + "| "
            + "| ".join([_normalize_cell(x, cell_width - 1) for x in row])
            + "|\n"
        )
        rst = rst + _table_div(num_cols, cell_width, header_flag)
        header_flag = 0
    return rst


def _create_fixer_list(all_fixers, grid):
    """
    Creates a list for the fixers table.
    Each fixer creates a list of its name, its min version and its max version if it exists.
    Ex: ["fix_deletion_http_request_HttpRequest_raw_post_data", "1.6", ""]
    """
    for fixer in all_fixers:
        format_tuple = (
            fixer["fixer_explanation"].replace("\n", "").strip(),
            fixer["fixer_callable"].__name__,
        )
        grid.append(
            [
                "**{}** (:code:`{}`)".format(*format_tuple),
                fixer["fixer_family"],
                detuplify_software_version(fixer["fixer_applied_from_version"]) or "",
                detuplify_software_version(fixer["fixer_applied_upto_version"]) or "",
            ]
        )


def _create_headers(grid):
    table_headers = ["Fixer and its ID", "Fixer family", "Min version", "Max version"]
    grid.append(table_headers)


def _make_rst_table(patching_registry):
    all_fixers = patching_registry.get_all_fixers()
    all_fixers.sort(key=lambda f: (f["fixer_reference_version"], f["fixer_id"]))
    grid = []
    _create_headers(grid=grid)
    _create_fixer_list(all_fixers, grid=grid)
    rst_table = make_table(grid=grid)
    return rst_table


def generate_readme(input_filename, output_filename, patching_registry):
    """
    Take an input file in RESTRUCTUREDTEXT format, append the table of available fixers to it,
    and generate the actual readme.
    """
    patching_registry.populate()

    with open(input_filename, mode="r", encoding="utf-8") as readme_manual:
        readme_manual_content = readme_manual.read()

    result = readme_manual_content + "\n\n"
    result += "Table of fixers\n===============\n\n"
    result += "There are currently {} available fixers.\n\n".format(
        len(patching_registry.get_all_fixers())
    )
    result += _make_rst_table(patching_registry)

    with open(output_filename, mode="w", encoding="utf-8") as readme_final:
        readme_final.write(result)



from contextlib import contextmanager
import shlex
import os
import sys
import subprocess
import datetime

from cookiecutter.utils import rmtree

SOURCE_ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

@contextmanager
def inside_dir(dirpath):
    """
    Execute code from inside the given directory
    :param dirpath: String, path of the directory the command is being run.
    """
    old_path = os.getcwd()
    try:
        os.chdir(dirpath)
        yield
    finally:
        os.chdir(old_path)


@contextmanager
def bake_in_temp_dir(cookies, **kwargs):
    """
    Delete the temporal directory that is created when executing the tests

    :param cookies: pytest_cookies.Cookies, cookie to be baked and its
    temporal files will be removed
    """
    result = cookies.bake(template=SOURCE_ROOT_DIR, **kwargs)
    try:
        yield result
    finally:
        rmtree(str(result.project))


def run_inside_dir(command, dirpath):
    """
    Run a command from inside a given directory, returning the exit status

    :param command: Command that will be executed
    :param dirpath: String, path of the directory the command is being run.
    """
    with inside_dir(dirpath):
        return subprocess.check_call(shlex.split(command))


def check_output_inside_dir(command, dirpath):
    "Run a command from inside a given directory, returning the command output"
    with inside_dir(dirpath):
        return subprocess.check_output(shlex.split(command))


def test_bake_with_defaults(cookies):
    with bake_in_temp_dir(cookies) as result:
        assert result.project.isdir()
        assert result.exit_code == 0
        assert result.exception is None

        found_toplevel_files = [f.basename for f in result.project.listdir()]
        assert 'setup.py' in found_toplevel_files
        assert 'tox.ini' in found_toplevel_files
        assert 'tests' in found_toplevel_files


def test_bake_and_run_tests(cookies):
    with bake_in_temp_dir(cookies) as result:
        assert result.project.isdir()
        run_inside_dir('python setup.py test', str(result.project)) == 0


def test_bake_withspecialchars_and_run_tests(cookies):
    """Ensure that a `full_name` with double quotes does not break setup.py"""
    with bake_in_temp_dir(cookies, extra_context={'full_name': 'name "quote" name'}) as result:
        assert result.project.isdir()
        run_inside_dir('python setup.py test', str(result.project)) == 0


def test_bake_with_apostrophe_and_run_tests(cookies):
    """Ensure that a `full_name` with apostrophes does not break setup.py"""
    with bake_in_temp_dir(cookies, extra_context={'full_name': "O'connor"}) as result:
        assert result.project.isdir()
        run_inside_dir('python setup.py test', str(result.project)) == 0


def test_using_pytest(cookies):
    with bake_in_temp_dir(cookies) as result:
        assert result.project.isdir()
        # Test the new pytest target
        run_inside_dir('pytest', str(result.project)) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
import sys

os.chdir(os.path.dirname(os.path.abspath(__file__)))  # security

from setuptools import setup, find_packages

def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read().strip()


readme_filename = "README.rst"
if not os.path.exists("README.rst"):
    raise RuntimeError("%s not found, have you run generate_readme.py first?" % readme_filename)


classifiers = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Intended Audience :: Information Technology
License :: OSI Approved :: MIT License
Natural Language :: English
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS :: MacOS X
"""


needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
setup_requires = ["pytest-runner"] if needs_pytest else []

packages = find_packages(where="src")

setup(
    name="{{ cookiecutter.project_name }}",
    version=read_file("VERSION"),
    author="{{ cookiecutter.full_name.replace('\"', '\\\"') }}",
    author_email='{{ cookiecutter.email }}',
    url='https://github.com/pakal/compat-patcher',
    license="MIT",
    platforms=["any"],
    description="{{ cookiecutter.project_short_description }}",
    classifiers=filter(None, classifiers.split("\n")),
    long_description=read_file(readme_filename),
    package_dir={"": "src"},
    packages=packages,
    install_requires=["compat-patcher-core"],
    extras_require={},
    setup_requires=setup_requires,
    tests_require=[
        # Beware, keep in sync with tox.ini
        "pytest",
        "pytest-cov",
    ],
)



import sys, six

from compat_patcher_core.import_proxifier import (
    install_import_proxifier,
    register_module_alias,
    _is_new_style_proxifier,
)


def test_import_proxifier():

    install_import_proxifier()
    install_import_proxifier()  # idempotent

    register_module_alias("json.tool_alias", real_name="json.tool")
    register_module_alias("mylogging", real_name="logging")
    register_module_alias("mylogging", real_name="logging")
    register_module_alias("infinite_recursion", real_name="infinite_recursion2")
    register_module_alias("infinite_recursion2", real_name="infinite_recursion")
    register_module_alias("unexisting_alias", real_name="unexisting_module")
    register_module_alias("noparent.newmodule", real_name="csv")

    fullname = "logging.handlers"
    # print ("ALREADY THERE:", fullname, fullname in sys.modules)

    import json.tool_alias

    assert sys.modules["json.tool_alias"]
    assert json.tool_alias

    fullname = "json.comments"
    # print ("ALREADY THERE:", fullname, fullname in sys.modules)

    fullname = "json.tool"
    # print ("ALREADY THERE:", fullname, fullname in sys.modules)

    import json.tool
    import json.tool

    # if _is_new_style_proxifier:
    # print ("JSON TOOL SPEC loader_state:", json.tool.__spec__.loader_state)

    assert json.tool_alias is json.tool, (json.tool_alias, json.tool)
    assert json.tool.__name__ == "json.tool", json.tool.__name__
    if _is_new_style_proxifier:
        assert json.tool.__spec__.origin == "alias", json.tool.__spec__.origin
        assert json.tool.__spec__.name == "json.tool_alias"
        assert json.tool.__spec__.loader_state["aliased_spec"].name == "json.tool"

    from mylogging import config

    assert config.dictConfig
    assert "logging.config" in sys.modules

    from mylogging.handlers import RotatingFileHandler
    from logging.handlers import RotatingFileHandler as RotatingFileHandlerOriginal

    # print("mylogging led to RotatingFileHandler", RotatingFileHandler)
    assert RotatingFileHandler is RotatingFileHandlerOriginal

    try:
        import infinite_recursion
    except RuntimeError:
        pass
    else:
        raise RuntimeError("Import should have led to infinite recursion")

    try:
        import unexisting_alias
    except ImportError as e:
        assert "unexisting_alias" in str(e), (str(e), vars(e))
        assert "unexisting_module" in str(e), str(e)
    else:
        raise RuntimeError("import error not raised for unexisting_alias")

    try:
        import unexisting_module
    except ImportError as e:
        assert "unexisting_alias" not in str(e), str(e)
        assert "unexisting_module" in str(e), str(e)
    else:
        raise RuntimeError("import error not raised for unexisting_module")

    try:
        import noparent.newmodule
    except ImportError as e:
        assert "No module named" in str(e), str(e)
    else:
        raise RuntimeError("import error not raised for noparent.newmodule")

    # We test_compatibility_with_other_custom_importers():
    # Old versions of lib crashed with AssertionError due to wrong module name "six.moves.urllib_parse" set by six._importer
    register_module_alias("my_six_urllib_parse_alias", real_name="six.moves.urllib.parse")
    import my_six_urllib_parse_alias
    # Re-overridden by our own importer on python3 only
    assert my_six_urllib_parse_alias.__name__ == "six.moves.urllib.parse" if six.PY3 else "six.moves.urllib_parse"
    assert my_six_urllib_parse_alias.urlencode(dict(name="h\xc3llo")) == "name=h%C3%83llo" if six.PY3 else "name=h%C3llo"



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from compat_patcher_core.utilities import (
    PatchingUtilities,
    detuplify_software_version,
    tuplify_software_version,
    WarningsProxy,
)

example_settings = {
    "logging_level": "INFO",
    "enable_warnings": True,
    "patch_injected_objects": True,
}

default_patch_marker = "__COMPAT_PATCHED__"


def test_patch_injected_object():

    patching_utilities = PatchingUtilities(example_settings)

    import dummy_module

    class MyTemplateResponse:
        pass

    patching_utilities.inject_class(
        dummy_module, "MyTemplateResponse", MyTemplateResponse
    )
    assert getattr(dummy_module.MyTemplateResponse, default_patch_marker) == True
    del dummy_module.MyTemplateResponse

    response = MyTemplateResponse()

    patching_utilities.inject_attribute(dummy_module, "_response_", response)
    assert getattr(dummy_module._response_, default_patch_marker) == True
    del dummy_module._response_

    def delete_selected():
        pass

    patching_utilities.inject_callable(dummy_module, "delete_selected", delete_selected)
    assert getattr(dummy_module.delete_selected, default_patch_marker) == True

    patching_utilities.inject_module("new_dummy_module", dummy_module)
    import new_dummy_module

    patching_utilities.inject_module("csv.mymodule", dummy_module)
    import csv.mymodule  # Submodules are OK too

    patching_utilities.inject_module("noparent.mymodule", dummy_module)
    with pytest.raises(ImportError, match="No module named"):
        import noparent.module  # Parent module "noparent" must exist first

    assert getattr(new_dummy_module, default_patch_marker) == True

    def mycallable(added):
        return 42 + added

    source_object = MyTemplateResponse()
    source_object.my_attr = mycallable
    target_object = MyTemplateResponse()
    patching_utilities.inject_callable_alias(
        target_object,
        "other_attr",
        source_object=source_object,
        source_attrname="my_attr",
    )
    assert getattr(target_object.other_attr, default_patch_marker) == True
    assert target_object.other_attr(added=2) == 44

    patching_utilities.inject_import_alias("newcsv", real_name="csv")
    from newcsv import DictReader

    del DictReader
    import newcsv

    assert newcsv.__name__ == "csv"


def test_patch_injected_objects_setting():

    patching_utilities = PatchingUtilities(example_settings)

    import dummy_module

    class MyMock(object):
        def method(self):
            return True

    my_mock = MyMock()
    my_mock2 = MyMock()

    patching_utilities.apply_settings(dict(patch_injected_objects="myattrname"))
    patching_utilities.inject_attribute(dummy_module, "new_stuff", my_mock)
    assert dummy_module.new_stuff.myattrname is True
    patching_utilities.inject_attribute(dummy_module, "new_stuff2", my_mock.method)
    assert not hasattr(
        dummy_module.new_stuff2, "myattrname"
    )  # Bound method has no __dict__

    patching_utilities.apply_settings(dict(patch_injected_objects=False))
    patching_utilities.inject_attribute(dummy_module, "other_stuff", my_mock2)
    assert not dummy_module.other_stuff.__dict__  # No extra marker


def test_enable_warnings_setting():
    from compat_patcher_core.utilities import stdlib_warnings

    patching_utilities = PatchingUtilities(example_settings)

    warnings_proxy = WarningsProxy()  # For now, goes direct lyto stdlib

    stdlib_warnings.simplefilter("always", Warning)

    with stdlib_warnings.catch_warnings(record=True) as w:
        patching_utilities.emit_warning("this feature is obsolete!", DeprecationWarning)
    assert len(w) == 1
    warning = w[0]
    assert "this feature is obsolete!" in str(warning.message)

    with stdlib_warnings.catch_warnings(record=True) as w:
        warnings_proxy.warn("this feature is obsolete too!", DeprecationWarning)
    assert len(w) == 1
    warning = w[0]
    assert "this feature is obsolete too!" in str(warning.message)

    warnings_proxy.set_patching_utilities(patching_utilities)
    with stdlib_warnings.catch_warnings(record=True) as w:
        warnings_proxy.warn("this feature is obsolete again!", DeprecationWarning)
    assert len(w) == 1
    warning = w[0]
    assert "this feature is obsolete again!" in str(warning.message)

    patching_utilities.apply_settings(dict(enable_warnings=False))

    with stdlib_warnings.catch_warnings(record=True) as w:
        patching_utilities.emit_warning("this feature is dead!", DeprecationWarning)
    assert len(w) == 0  # well disabled

    with stdlib_warnings.catch_warnings(record=True) as w:
        warnings_proxy.warn("this feature is obsolete again!", DeprecationWarning)
    assert len(w) == 0  # well disabled


def test_logging_level_setting(capsys):

    patching_utilities = PatchingUtilities(example_settings)

    patching_utilities.emit_log("<DEBUGGING1>", "DEBUG")
    patching_utilities.emit_log("<INFORMATION1>")  # default value

    out, err = capsys.readouterr()
    assert "<DEBUGGING1>" not in err
    assert "<INFORMATION1>" in err

    patching_utilities.apply_settings(dict(logging_level=None))

    patching_utilities.emit_log("<DEBUGGING2>", "DEBUG")
    patching_utilities.emit_log("<INFORMATION2>", "INFO")

    out, err = capsys.readouterr()
    assert "<DEBUGGING2>" not in err
    assert "<INFORMATION2>" not in err

    patching_utilities.apply_settings(dict(logging_level="DEBUG"))

    patching_utilities.emit_log("<DEBUGGING3>", "DEBUG")
    patching_utilities.emit_log("<INFORMATION3>", "INFO")

    out, err = capsys.readouterr()
    assert "<DEBUGGING3>" in err
    assert "<INFORMATION3>" in err


def test_version_tuplify_detuplify():
    assert tuplify_software_version((5, 0)) == (5, 0)
    assert tuplify_software_version("5.0") == (5, 0)
    assert tuplify_software_version(None) is None
    assert detuplify_software_version((5, 0)) == "5.0"
    assert detuplify_software_version("5.0") == "5.0"
    assert detuplify_software_version(None) is None



from __future__ import absolute_import, print_function, unicode_literals

from docutils.core import publish_parts

from compat_patcher_core.readme_generator import generate_readme
from dummy_fixers import patching_registry


def test_generate_readme(tmp_path):

    readme_in_file = tmp_path / "Readme.in"

    readme_in_file.write_text("THIS IS A README\n####################\n\n")

    readme_out_file = tmp_path / "Readme.txt"

    print(readme_in_file, readme_out_file)
    generate_readme(str(readme_in_file), str(readme_out_file), patching_registry)

    content = readme_out_file.read_text()

    html_body = publish_parts(source=content, writer_name="html4css1")["html_body"]

    assert "THIS IS A README" in html_body
    assert "available fixers" in html_body
    assert "<table" in html_body



from __future__ import absolute_import, print_function, unicode_literals

from compat_patcher_core.utilities import WarningsProxy

# Proxy meant to be imported on top of copy/pasted legacy code, instead of the stdlib "warnings" package.
# From inside fixers, it's best to use utils.emit_warnings() directly.
warnings = WarningsProxy()

# Here put DeprecationWarning subclasses and other deprecation utilities for your fixers



import os

import pytest
from compat_patcher_core.scaffolding import ensure_no_stdlib_warnings


def test_ensure_no_stdlib_warnings_in_package():
    import warnings  # This line will trigger checker error
    del warnings

    import {{ cookiecutter.project_slug }}

    pkg_root = os.path.dirname({{ cookiecutter.project_slug }}.__file__)
    analysed_files = ensure_no_stdlib_warnings(pkg_root)
    assert len(analysed_files) >= 3, analysed_files

    test_root = os.path.dirname(__file__)
    with pytest.raises(ValueError, match="wrong phrase.*test_scaffolding.py"):
        ensure_no_stdlib_warnings(test_root)


def test_no_package_shadowing_in_tox():
    import {{ cookiecutter.project_slug }}

    package_dir = os.path.dirname(os.path.abspath({{ cookiecutter.project_slug }}.__file__))
    if os.getenv("INSIDE_TOX_FOR_{{ cookiecutter.project_prefix.upper() }}") and ".tox" not in package_dir:
        raise RuntimeError("Wrong {{ cookiecutter.project_slug }} package used in Tox")



def ensure_no_stdlib_warnings(
    source_root,
    # we authorize "warnings.warn", as long as it uses the custom WarningsProxy above
    forbidden_phrases=(r"^\s*import.* warnings", r"^\s*from warnings"),
):
    """
    This utility should be used by each compat patcher user, to ensure all shims only
    go through the configurable compat-patcher warnings system.

    Returns the list of checked python source files.
    """
    import os, re

    analysed_files = []

    for root, _subdirs, files in os.walk(source_root):
        for f in [x for x in files if x.endswith(".py")]:
            full_path = os.path.join(root, f)
            # print(">> ANALYSING PYTHON FILE", full_path)
            with open(full_path, "r") as s:
                data = s.read()
            for forbidden_phrase in forbidden_phrases:
                if re.search(forbidden_phrase, data, re.MULTILINE):
                    if (f == "utilities.py") and (
                        "import " + "warnings as stdlib_warnings" in data
                    ):
                        continue  # the only case OK is our own warnings utility
                    raise ValueError(
                        "ALERT, wrong phrase '%s' detected in %s"
                        % (forbidden_phrase, full_path)
                    )
            analysed_files.append(f)
    return analysed_files


def ensure_all_fixers_have_a_test_under_pytest(
    config, items, patching_registry, _fail_fast=False
):
    """Call this pytest hook from a conftest.py to ensure your own test suite covers
    all your registered fixers, like so::

        def pytest_collection_modifyitems(config, items):
            from yourownpackage.registry import your_patching_registry
            from compat_patcher_core.scaffolding import ensure_all_fixers_have_a_test_under_pytest
            ensure_all_fixers_have_a_test_under_pytest(
                config=config, items=items, patching_registry=your_patching_registry
            )
    """
    import copy
    from _pytest.python import Function

    all_fixers = patching_registry.get_all_fixers()
    all_tests_names = [test.name for test in items]
    for fixer in all_fixers:
        expected_test_name = "test_{}".format(fixer["fixer_callable"].__name__)
        if expected_test_name not in all_tests_names:
            error_message = "No test written for {} fixer '{}'".format(
                fixer["fixer_family"].title(), fixer["fixer_callable"].__name__
            )

            def missing_fixer_test():
                raise RuntimeError(error_message)

            if _fail_fast:  # For testing only
                missing_fixer_test()
            mock_item = copy.copy(
                items[0]
            )  # We expect at least 1 test in the test suite, else it breaks...
            mock_item.parent.name = "test_{}_fixers.py".format(fixer["fixer_family"])
            setattr(
                mock_item.parent.obj,
                "MISSING_" + expected_test_name,
                missing_fixer_test,
            )

            parent = mock_item.parent
            params = dict(
                name="MISSING_" + expected_test_name,
                #config=config,
                #session=mock_item.session,
            )
            # See https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent
            if hasattr(Function, "from_parent"):
                function = Function.from_parent(parent, **params)
            else:
                function = Function(parent=parent, **params)

            items.append(function)



import os

import pytest

import compat_patcher_core
from compat_patcher_core import PatchingRegistry
from compat_patcher_core.scaffolding import (
    ensure_no_stdlib_warnings,
    ensure_all_fixers_have_a_test_under_pytest,
)


def test_ensure_all_fixers_have_a_test_under_pytest():
    class DummyNode:
        pass

    patching_registry = PatchingRegistry("dummyname")

    @patching_registry.register_compatibility_fixer(fixer_reference_version="10.1")
    def my_dummy_fixer(utils):
        "A help string"
        pass

    node = DummyNode()
    node.name = "test_my_dummy_fixer"
    items = [node]

    # No problem, all registered fixers have their test here
    ensure_all_fixers_have_a_test_under_pytest(
        config=None, items=items, patching_registry=patching_registry, _fail_fast=True
    )

    # Error when attempting to clone first item to report the missing test
    with pytest.raises(RuntimeError, match="No test written for .* fixer"):
        ensure_all_fixers_have_a_test_under_pytest(
            config=None, items=[], patching_registry=patching_registry, _fail_fast=True
        )


def test_ensure_no_stdlib_warnings_in_package():
    import warnings  # This line will trigger checker error

    del warnings
    pkg_root = os.path.dirname(compat_patcher_core.__file__)
    analysed_files = ensure_no_stdlib_warnings(pkg_root)
    assert len(analysed_files) > 5, analysed_files

    test_root = os.path.dirname(__file__)
    with pytest.raises(ValueError, match="wrong phrase.*test_scaffolding.py"):
        ensure_no_stdlib_warnings(test_root)


def test_no_package_shadowing_in_tox():
    import compat_patcher_core

    package_dir = os.path.dirname(os.path.abspath(compat_patcher_core.__file__))
    if os.getenv("INSIDE_TOX_FOR_CPC") and ".tox" not in package_dir:
        raise RuntimeError("Wrong compat_patcher_core package used in Tox")



import sys, os

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "src"))

from compat_patcher_core.readme_generator import generate_readme
from {{ cookiecutter.project_slug }}.registry import patching_registry


def generate_dcp_readme():
    generate_readme(
        input_filename="README.in",
        output_filename="README.rst",
        patching_registry=patching_registry,
    )


if __name__ == "__main__":
    generate_dcp_readme()
    print("DONE")



from __future__ import absolute_import, print_function, unicode_literals

import functools
import importlib
import logging
import sys
import types
import warnings as stdlib_warnings  # Do NOT import/use elsewhere than here!

import six


def tuplify_software_version(version):
    """
    Coerces the version string (if not None), to a version tuple.
    E.g. "1.7.0" becomes (1, 7, 0). No version suffix like "alpha" is expected.
    """
    if version is None:
        return version
    if isinstance(version, six.string_types):
        version = tuple(int(x) for x in version.split("."))
    assert len(version) <= 5, version
    assert all(isinstance(x, six.integer_types) for x in version), version
    return version


def detuplify_software_version(version):
    """
    Coerces the version tuple (if not None), to a version string.
    E.g. (1, 7, 0) becomes "1.7.0".
    """
    if version is None:
        return version
    if isinstance(version, (tuple, list)):
        version = ".".join(str(number) for number in version)
    assert isinstance(version, six.string_types)
    return version


class WarningsProxy(object):
    """An instance of this class acts as a replacement for the stdlib "warnings"
    package, but it relies on a PatchingUtilities instance as soon as this one
    is provided - thus making Warnings controllable by compat patcher settings.
    """

    _patching_utilities = None

    def set_patching_utilities(self, patching_utilities):
        assert isinstance(patching_utilities, PatchingUtilities), patching_utilities
        self._patching_utilities = patching_utilities

    def warn(self, *args, **kwargs):
        if self._patching_utilities:
            self._patching_utilities.emit_warning(*args, **kwargs)
        else:
            # Compat patcher is not yet configured
            stdlib_warnings.warn(*args, **kwargs)


class PatchingUtilities(object):
    """
    An instance of this class is provided as first argument to each compatibility fixer
    applied.

    It provides handy tools to monkey-patch the software environment, and emit logs and
    Warnings in a controllable way.

    For better forward-compatibility, please call injection utilities though keyword
    arguments, and not positional ones (due to python2.7 support, it's not
    enforced yet).
    """

    _logging_level = None
    _enable_warnings = False
    _patch_injected_objects = None

    settings_keys_used = ["logging_level", "enable_warnings", "patch_injected_objects"]

    def __init__(self, settings):
        # We force extraction of values, in case settings is a lazy instance
        # and not just a dict
        assert settings, settings
        settings = {name: settings[name] for name in self.settings_keys_used}

        self.apply_settings(settings)

    def apply_settings(self, settings):
        """This method can be called at runtime, mainly to alter the emission of logs
        and Warnings by fixers. it's possible to provide only a subset of settings, the
        others remaining as is.
        """
        if "logging_level" in settings:
            assert settings["logging_level"] is None or hasattr(
                logging, settings["logging_level"]
            ), settings["logging_level"]
            self._logging_level = settings["logging_level"]
        if "enable_warnings" in settings:
            assert settings["enable_warnings"] in (True, False), settings[
                "enable_warnings"
            ]
            self._enable_warnings = settings["enable_warnings"]
        if "patch_injected_objects" in settings:
            patch_injected_objects = settings["patch_injected_objects"]
            if patch_injected_objects is True:
                patch_injected_objects = "__COMPAT_PATCHED__"  # Default marker name
            assert not patch_injected_objects or isinstance(
                patch_injected_objects, six.string_types
            ), repr(patch_injected_objects)
            self._patch_injected_objects = patch_injected_objects

    @staticmethod
    def _is_simple_callable(obj):
        return isinstance(
            obj, (types.FunctionType, types.BuiltinFunctionType, functools.partial)
        )

    def _patch_injected_object(self, object_to_patch):
        """Mark shim object with a custom boolean attribute, to help identify it via
        introspection.

        Returns a boolean indicating whether the marking worked (some objects don't
        have a writable __dict__).
        """
        assert object_to_patch not in (True, False, None), object_to_patch
        if self._patch_injected_objects:
            try:
                setattr(object_to_patch, self._patch_injected_objects, True)
                return True
            except AttributeError:
                return False  # properties, bound methods and such can't be modified
        return None

    def emit_log(self, message, level="INFO"):
        """A logger printing to stderr, since at some stages of patching, logging is
        not yet setup.

        Log is only output if `level` is gerater or equal the current `logging_level`
        setting.
        """
        min_logging_level = self._logging_level
        if min_logging_level is None:
            return  # No logging at all
        if getattr(logging, level) < getattr(logging, min_logging_level):
            return
        full_message = "[DCP_%s] %s" % (level, message)
        print(full_message, file=sys.stderr)

    def emit_warning(self, message, category=DeprecationWarning, stacklevel=1):
        """Similar to "warnings.warn()" of the stdlib, but only emits the Warning if
        `enable_warnings` setting is True.
        """
        if self._enable_warnings:
            stdlib_warnings.warn(message, category, stacklevel + 1)

    def inject_attribute(self, target_object, target_attrname, attribute):
        """Inject an attribute into an object of any type (module, class, instance...).

        :param target_object: The object to patch
        :param target_attrname: The name given to the new attribute in the object to patch
        :param attribute: The attribute to inject, which must not be a callable
        """
        assert attribute is not None
        assert not self._is_simple_callable(attribute), attribute
        assert not isinstance(attribute, six.class_types), attribute

        self._patch_injected_object(attribute)
        setattr(target_object, target_attrname, attribute)

    def inject_callable(self, target_object, target_callable_name, patch_callable):
        """Inject a simple callable (not a class) into an object of any type (module, class, instance...).

        :param target_object: The object to patch
        :param target_callable_name: The name given to the new callable in the object to patch
        :param patch_callable: The callable to inject, which must be a callable, but not a class
        """
        assert self._is_simple_callable(patch_callable), patch_callable

        self._patch_injected_object(patch_callable)
        setattr(target_object, target_callable_name, patch_callable)

    def inject_callable_alias(
        self, target_object, target_attrname, source_object, source_attrname
    ):
        """
        Create and inject an alias for the source callable (not a class), which also
        triggers a deprecation warning when called.

        Returns the created alias callable.

        :param target_object: The object to patch
        :param target_attrname: The name of the callable on the target object
        :param source_object: The object from which to get the callable
        :param source_attrname: The name of the callable on the source object
        """
        source_callable = getattr(source_object, source_attrname)
        assert self._is_simple_callable(source_callable), source_callable

        @functools.wraps(source_callable)
        def wrapper(*args, **kwds):
            # we dunno if it's a backwards or forwards compatibility shim...
            self.emit_warning(
                "%s.%s, which is an alias for %s.%s, was called. One of these is deprecated."
                % (target_object, source_attrname, source_object, source_attrname),
                category=DeprecationWarning,
            )
            return source_callable(*args, **kwds)

        self._patch_injected_object(wrapper)
        setattr(target_object, target_attrname, wrapper)

        return wrapper

    def inject_class(self, target_object, target_klassname, klass):
        """Inject a class into an object of any type (module, class, instance...).

        :param target_object: The object to patch
        :param target_klassname: The name given to the new class in the object to patch
        :param klass: The class to inject
        """
        assert isinstance(klass, six.class_types), klass

        self._patch_injected_object(klass)
        setattr(target_object, target_klassname, klass)

    def inject_module(self, target_module_name, module):
        """Inject a module in sys.modules, under the selected dotted name.

        When injecting a submodule, its parent module(s) must exist too.

        :param target_module_name: The dotted name of the new module in sys.modules
        :param module: The module object to inject
        """
        target_module_name = str(target_module_name)  # Python2 compatibility
        assert isinstance(module, types.ModuleType), module
        assert sys.modules.get(target_module_name) is None

        self._patch_injected_object(module)

        sys.modules[target_module_name] = module

    def inject_import_alias(self, alias_name, real_name):
        """Create an import alias for the selected module.

        This doesn't directly patch sys.modules, but instead uses the imports hooks
        of python, so that "import <alias_name>" and "import <real_name>" both end up
        importing the same the (sub)module object.

        When the alias is a submodule, its parent module(s) must exist too.

        :param alias_name: The dotted name of the alias module
        :param real_name: The dotted name of the real module
        """

        from compat_patcher_core import import_proxifier

        import_proxifier.install_import_proxifier()  # idempotent activation
        import_proxifier.register_module_alias(
            alias_name=alias_name, real_name=real_name
        )


def _import_attribute_from_dotted_string(dotted_string):
    """Turns `mymodule.mysubmodule.my_attr` into the imported my_attr
    object, be it a class or an instance.
    """
    module_name, attr_name = dotted_string.rsplit(".", 1)
    module = importlib.import_module(module_name)
    attribute = getattr(module, attr_name)
    return attribute



#!/usr/bin/env python

import os
import sys

os.chdir(os.path.dirname(os.path.abspath(__file__)))  # security

from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

def read_file(fname):
    if not os.path.isabs(fname):
        fname = os.path.join(ROOT_DIR, fname)
    return open(os.path.join(os.path.dirname(__file__), fname)).read().strip()


classifiers = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Intended Audience :: Information Technology
License :: OSI Approved :: MIT License
Natural Language :: English
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS :: MacOS X
"""


needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
setup_requires = ["pytest-runner"] if needs_pytest else []

setup(
    name="compat-patcher-core",
    version=read_file("VERSION"),
    author="Pascal Chambon & others",
    author_email="pythoniks@gmail.com",
    url="https://github.com/pakal/compat-patcher-core",
    license="MIT",
    platforms=["any"],
    description="A patcher system to allow easy and lasting API compatibility.",
    classifiers=filter(None, classifiers.split("\n")),
    long_description=read_file("README.rst"),
    package_dir={"": "src"},
    packages=["compat_patcher_core"],
    install_requires=["six"],
    extras_require={
        "build_sphinx": ["sphinx", "sphinx_rtd_theme"],
        "run_pylint": ["pylint", "pylint-quotes"],
    },
    setup_requires=setup_requires,
    tests_require=[
        # Beware, keep in sync with requirements.txt
        "pytest",
        "pytest-cov",
        "docutils",
    ],
)



from __future__ import absolute_import, print_function, unicode_literals

import os

import _test_utilities


def test_fix_behaviour_submodule_element():
    import my_import_alias_for_csv
    del my_import_alias_for_csv  # All worked!



from __future__ import absolute_import, print_function, unicode_literals

import functools

import pytest

from compat_patcher_core.registry import PatchingRegistry, MultiPatchingRegistry
from dummy_fixers import patching_registry, patching_registry_bis, patching_registry_ter


def test_registry_populate():

    assert patching_registry.populate() is None  # Nothing done
    assert patching_registry.populate() is None  # Again

    registry = PatchingRegistry("dummyfamily", populate_callable=lambda registry: XYZ)
    with pytest.raises(NameError):
        registry.populate()

    registry = PatchingRegistry(
        "dummyfamily2", populate_callable=lambda registry: registry
    )
    assert registry.populate() is registry
    assert registry.populate() is None  # Idempotent


def test_get_relevant_fixer_ids():
    def log(msg):
        print(msg)

    get_relevant_fixer_ids = functools.partial(
        patching_registry.get_relevant_fixer_ids, log=log
    )

    assert patching_registry._get_current_software_version() == "5.1"

    fixer_ids_v5 = (
        get_relevant_fixer_ids()
    )  # current_software_version defaults to registry's value
    assert set(fixer_ids_v5) == set(
        [
            "fix_something_from_v4",
            "fix_something_from_v5",
            "fix_something_upto_v6",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )
    assert len(patching_registry.get_relevant_fixers()) == 5  # Same

    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0", qualified=True)
    assert set(fixer_ids) == set(
        [
            "dummy4.0|fix_something_from_v4",
            "dummy5.0|fix_something_from_v5",
            "dummy5.0|fix_something_upto_v6",
            "dummy5.0|fix_something_always",
            "dummy5.0|fix_something_but_skipped",
        ]
    )

    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0")
    assert set(fixer_ids) == set(
        [
            "fix_something_from_v4",
            "fix_something_from_v5",
            "fix_something_upto_v6",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )

    fixer_ids = get_relevant_fixer_ids(current_software_version="2.2")
    assert set(fixer_ids) == set(["fix_something_always", "fix_something_but_skipped"])

    fixer_ids = get_relevant_fixer_ids(current_software_version="10")
    assert set(fixer_ids) == set(
        [
            "fix_something_from_v5",
            "fix_something_from_v6",
            "fix_something_from_v7",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )

    fixer_ids = get_relevant_fixer_ids(current_software_version="6.0")
    assert set(fixer_ids) == set(
        [
            "fix_something_from_v4",
            "fix_something_from_v5",
            "fix_something_from_v6",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )  # But not "upto v6"

    fixer_settings = dict(
        include_fixer_ids=None,
        include_fixer_families=None,
        exclude_fixer_ids=None,
        exclude_fixer_families=None,
    )
    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0", **fixer_settings)
    assert not fixer_ids

    fixer_settings = dict(
        include_fixer_ids=[],
        include_fixer_families=["dummy4.0"],
        exclude_fixer_ids=[],
        exclude_fixer_families=[],
    )
    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0", **fixer_settings)
    assert set(fixer_ids) == set(["fix_something_from_v4"])

    fixer_settings = dict(
        include_fixer_ids="*",
        include_fixer_families=None,
        exclude_fixer_ids=[],
        exclude_fixer_families=[],
    )
    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0", **fixer_settings)
    assert set(fixer_ids) == set(fixer_ids_v5)

    fixer_settings = dict(
        include_fixer_ids=[],
        include_fixer_families="*",
        exclude_fixer_ids=None,
        exclude_fixer_families=[],
    )
    fixer_ids = get_relevant_fixer_ids(current_software_version="5.0", **fixer_settings)
    assert set(fixer_ids) == set(fixer_ids_v5)

    fixer_settings = dict(
        include_fixer_ids=["fix_something_from_v4"],
        include_fixer_families=["dummy5.0"],
        exclude_fixer_ids=["fix_something_upto_v6"],
        exclude_fixer_families=None,
    )
    fixer_ids = get_relevant_fixer_ids(
        current_software_version="5.5.4", **fixer_settings
    )
    assert set(fixer_ids) == set(
        [
            "fix_something_from_v4",
            "fix_something_from_v5",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )

    fixer_settings = dict(
        include_fixer_ids=["fix_something_from_v4"],
        include_fixer_families=["dummy5.0"],
        exclude_fixer_ids=["unexisting_id"],
        exclude_fixer_families=["dummy4.0"],
    )
    fixer_ids = get_relevant_fixer_ids(
        current_software_version="5.5.4", **fixer_settings
    )
    assert set(fixer_ids) == set(
        [
            "fix_something_from_v5",
            "fix_something_upto_v6",
            "fix_something_always",
            "fix_something_but_skipped",
        ]
    )

    fixer_settings = dict(
        include_fixer_ids=["dummy4.0|fix_something_from_v4"],
        include_fixer_families=["dummy5.0"],
        exclude_fixer_ids=["dummy5.0|fix_something_always"],
    )
    fixer_ids = get_relevant_fixer_ids(
        current_software_version="5.5.4", qualified=True, **fixer_settings
    )
    assert set(fixer_ids) == set(
        [
            "dummy4.0|fix_something_from_v4",
            "dummy5.0|fix_something_from_v5",
            "dummy5.0|fix_something_upto_v6",
            "dummy5.0|fix_something_but_skipped",
        ]
    )

    fixer_settings = dict(
        include_fixer_ids=["fix_something_from_v4"],
        include_fixer_families=["dummy5.0"],
        exclude_fixer_ids=[],
        exclude_fixer_families="*",
    )
    fixer_ids = get_relevant_fixer_ids(
        current_software_version="5.5.4", **fixer_settings
    )
    assert fixer_ids == []

    fixer_settings = dict(
        include_fixer_ids="*",
        include_fixer_families=["dummy5.0"],
        exclude_fixer_ids="*",
        exclude_fixer_families=None,
    )
    fixer_ids = get_relevant_fixer_ids(
        current_software_version="5.5.4", **fixer_settings
    )
    assert fixer_ids == []


def test_get_fixer_by_id():
    res = patching_registry.get_fixer_by_id("fix_something_from_v7")
    assert isinstance(res, dict)
    assert res["fixer_id"] == "fix_something_from_v7"

    with pytest.raises(KeyError):
        patching_registry.get_fixer_by_id("ddssdfsdfsdf")


def test_get_all_fixers():
    res = patching_registry.get_all_fixers()
    assert len(res) == 7


def test_docstring_mandatory_on_fixers():
    with pytest.raises(ValueError):

        @patching_registry.register_compatibility_fixer(
            fixer_reference_version="5.0", fixer_applied_from_version="6.0"
        )
        def fixer_without_docstring(utils):
            pass


def test_registry_current_software_version():
    registry = PatchingRegistry(
        family_prefix="dummy2", current_software_version=lambda: AAA
    )
    with pytest.raises(NameError):
        registry._get_current_software_version()

    registry = PatchingRegistry(
        family_prefix="dummy3", current_software_version=lambda: (1, 2, 3)
    )
    assert registry._get_current_software_version() == (1, 2, 3)

    registry = PatchingRegistry(
        family_prefix="dummy4", current_software_version="8.2.1"
    )
    assert registry._get_current_software_version() == "8.2.1"

    registry = PatchingRegistry(family_prefix="dummy5")
    assert registry._get_current_software_version() is None
    with pytest.raises(ValueError, match="valid current_software_version"):
        registry.get_relevant_fixers()
    assert registry.get_relevant_fixers(current_software_version="1") == []


def test_multi_patching_registry():

    multi_registry = MultiPatchingRegistry(
        registries=[patching_registry, patching_registry_bis]
    )
    assert not patching_registry_bis._is_populated
    assert not multi_registry._is_populated

    assert len(patching_registry.get_all_fixers()) == 7
    assert len(patching_registry_bis.get_all_fixers()) == 3

    fixers = multi_registry.get_all_fixers()
    assert len(fixers) == 10  # All, including duplicate names!
    assert not patching_registry_bis._is_populated
    assert not multi_registry._is_populated

    fixer = multi_registry.get_fixer_by_id("fix_something_always")
    assert fixer["fixer_reference_version"] == (5, 0)  # First one is returned!
    with pytest.raises(KeyError):
        multi_registry.get_fixer_by_id("badname")
    assert not patching_registry_bis._is_populated
    assert not multi_registry._is_populated

    selected_fixers = multi_registry.get_relevant_fixer_ids()
    assert selected_fixers == [
        "fix_something_from_v4",
        "fix_something_from_v5",
        "fix_something_upto_v6",
        "fix_something_always",
        "fix_something_but_skipped",
        "fix_something_other_taken",
        "fix_something_always",
    ]  # Duplicate fixer is taken too
    assert patching_registry_bis._is_populated
    assert multi_registry._is_populated

    misc_fixers = multi_registry.get_relevant_fixer_ids(current_software_version="1000")
    assert misc_fixers == [
        "fix_something_from_v5",
        "fix_something_always",
        "fix_something_but_skipped",
        "fix_something_from_v6",
        "fix_something_from_v7",
        "fix_something_other_taken",
        "fix_something_other_not_taken",
        "fix_something_always",
    ]

    multi_registry2 = MultiPatchingRegistry(
        registries=[
            "dummy_fixers.patching_registry_bis",
            "dummy_fixers.patching_registry_ter_as_callable",
        ]
    )
    assert not multi_registry2._is_populated
    assert not patching_registry_ter._is_populated

    multi_registry2.populate()
    assert multi_registry2._is_populated
    assert patching_registry_ter._is_populated

    selected_fixers2 = multi_registry2.get_relevant_fixer_ids()

    assert selected_fixers2 == [
        "fix_something_other_taken",
        "fix_something_always",
        "fix_everything",
    ]

    with pytest.raises(ValueError, match="Wrong registry reference"):
        MultiPatchingRegistry(registries=["dummy_fixers.SkipFixerException"])

    with pytest.raises(ImportError):
        MultiPatchingRegistry(registries=["badmodule.unexisting"])

    with pytest.raises(AttributeError):
        MultiPatchingRegistry(registries=["dummy_fixers.unexisting"])

    with pytest.raises(ValueError, match="unpack"):
        MultiPatchingRegistry(registries=["badstring"])



import os

import compat_patcher_core
import dummy_module
from compat_patcher_core.exceptions import SkipFixerException
from compat_patcher_core.registry import PatchingRegistry

# Important, when testing with Tox, we must ensure no shadowing occurs
PACKAGE_DIR = os.path.dirname(os.path.abspath(compat_patcher_core.__file__))
print(
    '### TESTING DUMMY FIXERS WITH THE COMPAT_PATCHER_CORE FROM "%s" ###' % PACKAGE_DIR
)


patching_registry = PatchingRegistry(
    family_prefix="dummy", current_software_version=lambda: "5.1"
)


@patching_registry.register_compatibility_fixer(
    fixer_reference_version="4.0",
    fixer_applied_from_version="4.0",
    fixer_applied_upto_version="9.3",
    fixer_tags=[],
)
def fix_something_from_v4(utils):
    "Does something here"
    dummy_module.APPLIED_FIXERS.append(fix_something_from_v4.__name__)
    return 22


assert fix_something_from_v4(utils=33) == 22  # Decorator doesn't erase wrapped function


@patching_registry.register_compatibility_fixer(
    fixer_reference_version="5.0",
    fixer_applied_from_version="5.0",
    fixer_tags=["mytag"],
)
def fix_something_from_v5(utils):
    "Does something there"
    dummy_module.APPLIED_FIXERS.append(fix_something_from_v5.__name__)


@patching_registry.register_compatibility_fixer(
    fixer_reference_version="5.0",
    fixer_applied_from_version="4.2",
    fixer_applied_upto_version="6.0",
)
def fix_something_upto_v6(utils):
    "Does something again"
    dummy_module.APPLIED_FIXERS.append(fix_something_upto_v6.__name__)


@patching_registry.register_compatibility_fixer(fixer_reference_version="5.0")
def fix_something_always(utils):
    "Does something always"
    dummy_module.APPLIED_FIXERS.append(fix_something_always.__name__)


@patching_registry.register_compatibility_fixer(fixer_reference_version="5.0")
def fix_something_but_skipped(utils):
    "Does something always"
    raise SkipFixerException("Fixer fix_something_but_skipped is always skipped")


@patching_registry.register_compatibility_fixer(
    fixer_reference_version="6.0",
    fixer_applied_from_version="6.0",
    feature_supported_upto_version="6.5",
)
def fix_something_from_v6(utils):
    "Does something bis"
    dummy_module.APPLIED_FIXERS.append(fix_something_from_v6.__name__)


@patching_registry.register_compatibility_fixer(
    fixer_reference_version="6.3",
    fixer_applied_from_version="7.3",
    feature_supported_from_version="5.0",
    feature_supported_upto_version="11.2",
)
def fix_something_from_v7(utils):
    "Does something still"
    dummy_module.APPLIED_FIXERS.append(fix_something_from_v7.__name__)


### -------------------

patching_registry_bis = PatchingRegistry(
    family_prefix="other", current_software_version=lambda: "10"
)


@patching_registry_bis.register_compatibility_fixer(
    fixer_reference_version="6.3", fixer_applied_from_version="7.3"
)
def fix_something_other_taken(utils):
    "Does something still"
    dummy_module.APPLIED_FIXERS.append(fix_something_other_taken.__name__)


@patching_registry_bis.register_compatibility_fixer(
    fixer_reference_version="6.3", fixer_applied_from_version="20"
)
def fix_something_other_not_taken(utils):
    "Does something still"
    dummy_module.APPLIED_FIXERS.append(fix_something_other_not_taken.__name__)


assert fix_something_always  # Already existing


def __fix_something_always(utils):
    "Does something still"
    dummy_module.APPLIED_FIXERS.append(__fix_something_always.__name__)


__fix_something_always.__name__ = "fix_something_always"

patching_registry_bis.register_compatibility_fixer(fixer_reference_version="8.3")(
    __fix_something_always
)

### -------------------

patching_registry_ter = PatchingRegistry(
    family_prefix="somefamily", current_software_version=lambda: "100"
)


@patching_registry_ter.register_compatibility_fixer(fixer_reference_version="8.3")
def fix_everything(utils):
    "Does something still"
    dummy_module.APPLIED_FIXERS.append(fix_everything.__name__)


patching_registry_ter_as_callable = lambda: patching_registry_ter



# -*- coding: utf-8 -*-


from compat_patcher_core import generic_patch_software, make_safe_patcher, DEFAULT_SETTINGS


@make_safe_patcher
def patch(settings=None):

    # TODO replace this by default settings extracted from the patched framework
    settings = settings if settings is not None else DEFAULT_SETTINGS

    from .registry import patching_registry
    # Use the following import form to avoid triggering checker alerts on "warnings" import
    import {{ cookiecutter.project_slug }}.deprecation

    generic_patch_software(
        settings=settings,
        patching_registry=patching_registry,
        warnings_proxy=deprecation.warnings,
    )

